import os
import logging

from gemini import GeminiQuery

from puzzle.models import (Case, Individual)

from puzzle.plugins import BaseCaseMixin

logger = logging.getLogger(__name__)

class CaseMixin(BaseCaseMixin):
    """Class to store methods that deal with Cases in geimni plugin"""
    
    def _add_individual(self, ind_obj):
        """Add a individual to the adapter
        
            Args:
                ind_obj (puzzle.models.Individual)
        """
        logger.debug("Adding individual {0} to plugin".format(ind_obj.ind_id))
        self.individual_objs.append(ind_obj)

    def add_case(self, case_obj):
        """Add a case obj with individuals to adapter
        
            Args:
                case_obj (puzzle.models.Case)
                
        """
        for ind_obj in case_obj.individuals:
            self._add_individual(ind_obj)
        logger.debug("Adding case {0} to plugin".format(case_obj.case_id))
        self.case_objs.append(case_obj)

    def cases(self, pattern=None):
        """Return all cases.
        
            Args:
                pattern: Allways None in gemini adapter
            
            Returns:
                case_objs(An iterator with Cases)
        """

        return self.case_objs

    def case(self, case_id=None):
        """Return a Case object

            If no case_id is given return one case

            Args:
                case_id (str): A case id

            Returns:
                case(Case): A Case object
        """
        cases = self.cases()
        if case_id:
            for case in cases:
                if case.case_id == case_id:
                    return case
        else:
            if cases:
                return cases[0]

        return None

    def individual(self, ind_id=None):
        """Return a individual object
        
            Args:
                ind_id (str): A individual id
            
            Returns:
                individual (puzzle.models.individual)
        """
        for ind_obj in self.individuals:
            if ind_obj.ind_id == ind_id:
                return ind_obj
        return None

    def individuals(self, *ind_ids):
        """Return information about individuals
        
            Args:
                ind_ids (list(str)): List of individual ids
            
            Returns:
                individuals (Iterable): Iterable with Individuals
        """
        if ind_ids:
            for ind_id in ind_ids:
                for ind in self.individual_objs:
                    if ind.ind_id == ind_id:
                        yield ind
        else:
            for ind in self.individual_objs:
                yield ind



import logging

from vcftoolbox import Genotype

from puzzle.models import Genotype as puzzle_genotype

logger = logging.getLogger(__name__)

class GenotypeExtras(object):
    """Class to store methods that deals with genotyping"""

    def _add_genotype_calls(self, variant_obj, variant_line, case_obj):
        """Add the genotype calls for the variant

        Args:
            variant_obj (puzzle.models.Variant)
            variant_dict (dict): A variant dictionary
            case_obj (puzzle.models.Case)

        """
        variant_line = variant_line.split('\t')
        #if there is gt calls we have no individuals to add
        if len(variant_line) > 8:
            gt_format = variant_line[8].split(':')
            for individual in case_obj.individuals:
                sample_id = individual.ind_id
                index = individual.ind_index

                gt_call = variant_line[9+index].split(':')

                raw_call = dict(zip(gt_format, gt_call))

                genotype = Genotype(**raw_call)

                variant_obj.add_individual(puzzle_genotype(
                    sample_id = sample_id,
                    genotype = genotype.genotype,
                    case_id = case_obj.name,
                    phenotype = individual.phenotype,
                    ref_depth = genotype.ref_depth,
                    alt_depth = genotype.alt_depth,
                    genotype_quality = genotype.genotype_quality,
                    depth = genotype.depth_of_coverage,
                    supporting_evidence = genotype.supporting_evidence,
                    pe_support = genotype.pe_support,
                    sr_support = genotype.sr_support,
                ))



# -*- coding: utf-8 -*-
from .dotdict import DotDict


class Case(DotDict):
    """Case representation."""
    def __init__(self, case_id, name=None, variant_source=None,
                 variant_type='snv', variant_mode='vcf', compressed=False,
                 tabix_index=False):
        super(Case, self).__init__(
            case_id=case_id,
            name=name or case_id,
            variant_source=variant_source,
            variant_type=variant_type,
            variant_mode=variant_mode,
            compressed=compressed,
            tabix_index=tabix_index,
        )

        self['individuals'] = []

    def add_individual(self, individual):
        """Add a individual object to the case

            Args:
                individual (dict): An Individual object
        """
        self['individuals'].append(individual)



from puzzle.utils.constants import HGNC_TO_OMIM, SEVERITY_DICT

def test_HGNC_TO_OMIM():
    assert HGNC_TO_OMIM['CACNA1F'].get('mim_nr') == 300110
    assert HGNC_TO_OMIM['ADK'].get('mim_nr') == 102750

def test_SEVERITY_DICT():
    assert SEVERITY_DICT['transcript_ablation'] == 0
    assert SEVERITY_DICT['start_lost'] == 6



# -*- coding: utf-8 -*-
import logging
import hashlib
from . import (DotDict, Transcript, Gene, Compound, Genotype)

logger = logging.getLogger(__name__)


class Variant(DotDict):
    """docstring for Variant"""
    def __init__(self, CHROM, POS, ID, REF, ALT, QUAL, FILTER):
        super(Variant, self).__init__(CHROM=CHROM, POS=POS, ID=ID, REF=REF,
                                      ALT=ALT, QUAL=QUAL, FILTER=FILTER)

        self._set_variant_id()
        self['index'] = None

        self['thousand_g'] = None  # float
        self['max_freq'] = None  # float
        self['cadd_score'] = None  # float
        self['consequences'] = []
        self['most_severe_consequence'] = None  # str
        self['impact_severities'] = []  # List of severities
        self['rank_score'] = None  # float

        self['frequencies'] = []
        self['severities'] = []
        self['transcripts'] = []  # List of Transcripts
        self['individuals'] = []  # List of Genotypes
        self['genes'] = []  # List of Genes
        self['gene_symbols'] = []  # List of gene symbols
        self['compounds'] = []  # List of Compounds
        self['genetic_models'] = []  # List of genetic models followed
        #SV specific fields:
        self['sv_type'] = None
        self['sv_len'] = None
        self['stop_chrom'] = None
        self['stop'] = None
        self['cytoband_start'] = None
        self['cytoband_stop'] = None
        self['consequences'] = []

    @property
    def nr_genes(self):
        """Return the number of genes"""
        if self['genes']:
            nr_genes = len(self['genes'])
        else:
            nr_genes = len(self['gene_symbols'])
        return nr_genes

    @property
    def is_snv(self):
        """Check if variant is SNV or SV."""
        return self.get('cytoband_start') is None

    @property
    def display_name(self):
        """Readable name for the variant."""
        if self.is_snv:
            gene_ids = self.gene_symbols[:2]
            return ', '.join(gene_ids)
        else:
            return "{this.cytoband_start} ({this.sv_len})".format(this=self)

    @property
    def md5(self):
        """Return a md5 key string based on position, ref and alt"""
        return hashlib.md5('_'.join([self.CHROM, str(self.POS), self.REF,
                                     self.ALT])).hexdigest()

    @property
    def is_intrachromosomal(self):
        """Check if variant is intrachromosomal

            If stop_chrom == CHROM return True
            else return False
        """
        return self.get('stop_chrom', self['CHROM']) == self['CHROM']

    def add_frequency(self, name, value):
        """Add a frequency that will be displayed on the variant level

            Args:
                name (str): The name of the frequency field
        """
        logger.debug("Adding frequency {0} with value {1} to variant {2}".format(
            name, value, self['variant_id']))
        self['frequencies'].append({'label': name, 'value': value})

    def set_max_freq(self, max_freq=None):
        """Set the max frequency for the variant

            If max_freq use this, otherwise go through all frequencies and
            set the highest as self['max_freq']

            Args:
                max_freq (float): The max frequency
        """
        if max_freq:
            self['max_freq'] = max_freq
        else:
            for frequency in self['frequencies']:
                if self['max_freq']:
                    if frequency['value'] > self['max_freq']:
                        self['max_freq'] = frequency['value']
                else:
                    self['max_freq'] = frequency['value']
        return

    def add_severity(self, name, value):
        """Add a severity to the variant

            Args:
                name (str): The name of the severity
                value : The value of the severity
        """
        logger.debug("Adding severity {0} with value {1} to variant {2}".format(
            name, value, self['variant_id']))
        self['severities'].append({name: value})

    def add_individual(self, genotype):
        """Add the information for a individual

            This adds a genotype dict to variant['individuals']

            Args:
                genotype (dict): A genotype dictionary
        """
        logger.debug("Adding genotype {0} to variant {1}".format(
            genotype, self['variant_id']))
        self['individuals'].append(genotype)

    def add_transcript(self, transcript):
        """Add the information transcript

            This adds a transcript dict to variant['transcripts']

            Args:
                transcript (dict): A transcript dictionary
        """
        logger.debug("Adding transcript {0} to variant {1}".format(
            transcript, self['variant_id']))
        self['transcripts'].append(transcript)

    def add_gene(self, gene):
        """Add the information of a gene

            This adds a gene dict to variant['genes']

            Args:
                gene (dict): A gene dictionary

        """
        logger.debug("Adding gene {0} to variant {1}".format(
            gene, self['variant_id']))
        self['genes'].append(gene)

    def add_compound(self, compound):
        """Add the information of a compound variant

            This adds a compound dict to variant['compounds']

            Args:
                compound (dict): A compound dictionary

        """
        logger.debug("Adding compound {0} to variant {1}".format(
            compound, self['variant_id']))
        self['compounds'].append(compound)

    def update_variant_id(self, variant_id):
        """Update the variant id for an individual

            Args:
                variant_id (str): A variant id
        """

        self._set_variant_id(variant_id=variant_id)

    def _set_variant_id(self, variant_id=None):
        """Set the variant id for this variant"""
        if not variant_id:
            variant_id = '_'.join([
                self.CHROM,
                str(self.POS),
                self.REF,
                self.ALT
                ])

        logger.debug("Updating variant id to {0}".format(
            variant_id))

        self['variant_id'] = variant_id

    def __repr__(self):
        return ("Variant(CHROM={this.CHROM},POS={this.POS},REF={this.REF},ALT={this.ALT})"
                        .format(this=self))



# -*- coding: utf-8 -*-
import logging

from puzzle.models.dotdict import DotDict
from puzzle.utils import get_gene_info

logger = logging.getLogger(__name__)


class Plugin(object):
    """docstring for Plugin"""
    def __init__(self):
        super(Plugin, self).__init__()
        self.db = None
        self.puzzle_db = None
        self.individual_objs = None
        self.case_objs = None
        self.variant_type = 'snv'
        self.filters = DotDict(
            can_filter_frequency=False,
            can_filter_cadd=False,
            can_filter_consequence=False,
            can_filter_gene=False,
            can_filter_inheritance=False,
            can_filter_sv=False,
            can_filter_impact_severity=False,
            can_query_gemini=False,
            can_filter_range=False,
        )

    def init_app(self, app):
        """Initialize plugin via Flask."""
        raise NotImplementedError

    def individual_dict(self, ind_ids):
        """Return a dict with ind_id as key and Individual as values."""
        ind_dict = {ind.ind_id: ind for ind in self.individuals(ind_ids=ind_ids)}
        return ind_dict


class GeneInfoMixin(object):
    def _get_genes(self, variant):
        """Add the genes for a variant

        Get the hgnc symbols from all transcripts and add them
        to the variant.

        Args:
            variant (dict): A variant dictionary

        Returns:
            genes (list): A list of Genes
        """
        transcripts = variant['transcripts']
        ensembl_ids = [transcript['ensembl_id'] for transcript in
                       transcripts if transcript['ensembl_id']]
        hgnc_symbols = [transcript['hgnc_symbol'] for transcript in
                        transcripts if transcript['hgnc_symbol']]
        genes = get_gene_info(ensembl_ids, hgnc_symbols)
        return genes



# -*- coding: utf-8 -*-
import logging

from flask import Flask

from .ext import bootstrap, markdown
from .settings import BaseConfig

logger = logging.getLogger(__name__)


def create_app(config=None, config_obj=None):
    """Flask app factory function.

    Args:
        config (Optional[path]): path to a Python module config file
        config_obj (Optional[class]): Python config object
    """
    app = Flask(__name__)
    # configure application from external configs
    configure_app(app, config=config, config_obj=config_obj)
    # register different parts of the application
    register_blueprints(app)
    # setup extensions
    bind_extensions(app)
    return app


def configure_app(app, config=None, config_obj=None):
    """Configure application instance.

    Args:
        app (Flask): initialized Flask app instance
        config (Optional[path]): path to a Python module config file
        config_obj (Optional[class]): Python config object
    """
    app.config.from_object(config_obj or BaseConfig)
    if config is not None:
        app.config.from_pyfile(config)


def register_blueprints(app):
    """Configure blueprints.

    Args:
        app (Flask): initialized Flask app instance
    """
    for blueprint in app.config.get('BLUEPRINTS', []):
        app.register_blueprint(blueprint)


def bind_extensions(app):
    """Configure extensions.

    Args:
        app (Flask): initialized Flask app instance
    """
    # bind plugin to app object
    app.db = app.config['PUZZLE_BACKEND']
    app.db.init_app(app)

    # bind bootstrap blueprints
    bootstrap.init_app(app)
    markdown(app)

    @app.template_filter('islist')
    def islist(object):
        return isinstance(object, (tuple, list))



from .models import Case, Individual, BASE
from .phenotypeterm import PhenotypeTerm
from .genelist import GeneList, CaseGenelistLink
from .resource import Resource
from .query import GeminiQuery
from .comment import Comment
from .suspect import Suspect



import click
import puzzle

root = click.option('--root', '-r',
                        type=click.Path(),
                        help="Path to where to find variant source(s)"
                    )

mode = click.option('-m', '--mode',
                        type=click.Choice(['vcf', 'gemini']),
                        default='vcf',
                        show_default=True
                        )

verbose = click.option('-v', '--verbose',
                        count=True,
                        default=2
                        )

variant_type = click.option('--variant-type',
                type=click.Choice(['snv', 'sv']),
                default='snv',
                show_default=True,
                help="If Structural Variantion or Single Nucleotide variant"\
                " mode should be used"
                )
family_file = click.option('-f', '--family_file',
                type=click.File('r')
                )
family_type = click.option('--family_type',
                type=click.Choice(['ped', 'alt']),
                default='ped',
                show_default=True,
                help='If the analysis use one of the known setups, please specify which one.'
)
version = click.version_option(puzzle.__version__)

phenomizer = click.option('--phenomizer', 
                nargs=2, 
                help='Phenomizer username/password',
                envvar='PHENOMIZER_AUTH'
                )


# -*- coding: utf-8 -*-
import hashlib

def test_variant(variant):
    """docstring for test_variant"""
    assert variant['CHROM'] == '1'
    assert variant['variant_id'] == '1_100_A_T'


def test_adding_transcript(variant):
    """docstring for test_variant"""
    assert variant['transcripts'] == []

    transcript = {'SYMBOL': 'ADK'}
    variant.add_transcript(transcript)

    assert variant['transcripts'] == [transcript]


def test_add_frequency(variant):
    """docstring for test_variant"""
    assert variant['frequencies'] == []
    variant.add_frequency(name='1000G', value=0.01)
    assert variant['frequencies'] == [{'label': '1000G', 'value': 0.01}]


def test_add_severity(variant):
    """docstring for test_variant"""

    assert variant['severities'] == []

    variant.add_severity(
        name='SIFT',
        value=0.9
    )

    assert variant['severities'] == [{'SIFT':0.9}]


def test_add_individual(variant):
    """docstring for test_variant"""

    assert variant['individuals'] == []

    genotype = {'GT':'0/1'}

    variant.add_individual(genotype)

    assert variant['individuals'] == [genotype]


def test_add_gene(variant):
    """docstring for test_variant"""

    assert variant['genes'] == []

    gene = {'HGNC_ID':'ADK'}

    variant.add_gene(gene)

    assert variant['genes'] == [gene]


def test_add_compound(variant):
    """docstring for test_variant"""

    assert variant['compounds'] == []

    compound = {'variant_id':'1_2_A_C'}

    variant.add_compound(compound)

    assert variant['compounds'] == [compound]


def test_update_variant_id(variant):
    """docstring for test_variant"""

    assert variant['variant_id'] == '1_100_A_T'

    variant_id = 'family1_1_2_A_C'

    variant.update_variant_id(variant_id)

    assert variant['variant_id'] == variant_id

def test_md5(variant):
    id_string = '_'.join([variant.CHROM, variant.POS, variant.REF, variant.ALT])
    md5_string = hashlib.md5(id_string).hexdigest()
    assert variant.md5 == md5_string



# -*- coding: utf-8 -*-


def test_add_phenotype(test_db, phenomizer_auth):
    # setup phenomizer auth
    test_db.phenomizer_auth = phenomizer_auth

    ind_id = 'ADM1059A2'
    ind_obj = test_db.individual(ind_id)

    hpo_id = 'HP:0002497'
    added_terms = test_db.add_phenotype(ind_obj, hpo_id)

    assert len(added_terms) == 1
    assert len(ind_obj.phenotypes) == 1

    term_obj = ind_obj.phenotypes[0]
    assert term_obj.phenotype_id == hpo_id
    assert term_obj.description == 'Spastic ataxia'
    assert term_obj.individual.ind_id == ind_id

    # test with a missing HPO id
    hpo_id = 'HP:0000118'
    assert test_db.add_phenotype(ind_obj, hpo_id) is None

    # test with a repeated HPO id
    hpo_id = 'HP:0002497'
    assert test_db.add_phenotype(ind_obj, hpo_id) == []

    # test adding with OMIM id
    omim_id = 'OMIM:615862'
    hpo_ids = ['HP:0000092', 'HP:0006580', 'HP:0000090', 'HP:0000238',
               'HP:0001249', 'HP:0001970', 'HP:0004722', 'HP:0000486',
               'HP:0003774', 'HP:0001396']
    added_terms = test_db.add_phenotype(ind_obj, omim_id)
    added_ids = [term.phenotype_id for term in added_terms]
    assert set(added_ids) == set(hpo_ids)
    assert len(ind_obj.phenotypes) == len(hpo_ids) + 1


def test_remove_phenotype(test_db, phenomizer_auth):
    # setup phenomizer auth
    test_db.phenomizer_auth = phenomizer_auth

    ind_id = 'ADM1059A2'
    ind_obj = test_db.individual(ind_id)

    hpo_id = 'HP:0002497'
    removed_hpo_id = 'HP:0006580'
    test_db.add_phenotype(ind_obj, hpo_id)
    test_db.add_phenotype(ind_obj, removed_hpo_id)

    assert len(ind_obj.phenotypes) == 2
    test_db.remove_phenotype(ind_obj, removed_hpo_id)
    assert len(ind_obj.phenotypes) == 1
    assert ind_obj.phenotypes[0].phenotype_id == hpo_id

    # test resetting the phenotypes for an individual
    test_db.remove_phenotype(ind_obj)
    assert len(ind_obj.phenotypes) == 0



import logging

from vcftoolbox import (get_vcf_handle,get_variant_id)

from cyvcf2 import VCF

from puzzle.plugins import BaseVariantMixin
from puzzle.plugins.constants import Results

from puzzle.models import (Variant)

from puzzle.utils import (get_most_severe_consequence, get_omim_number,
                          get_csq, IMPACT_SEVERITIES, get_header)

from .variant_extras import VariantExtras

logger = logging.getLogger(__name__)

class VariantMixin(BaseVariantMixin, VariantExtras):
    """Class to store variant specific functions for vcf plugin"""

    def variant(self, case_id, variant_id):
        """Return a specific variant.

            Args:
                case_id (str): Path to vcf file
                variant_id (str): A variant id

            Returns:
                variant (Variant): The variant object for the given id
        """
        case_obj = self.case(case_id=case_id)
        vcf_file_path = case_obj.variant_source
        self.head = get_header(vcf_file_path)

        self.vep_header = self.head.vep_columns
        self.snpeff_header = self.head.snpeff_columns

        handle = VCF(vcf_file_path)

        for index, variant in enumerate(handle):
            index += 1
            line_id = get_variant_id(variant_line=str(variant)).lstrip('chrCHR')
            if line_id == variant_id:
                return self._format_variants(
                    variant=variant,
                    index=index,
                    case_obj=case_obj,
                    add_all_info=True
                    )

        return None

    def variants(self, case_id, skip=0, count=1000, filters=None):
        """Return all variants in the VCF.

        This function will apply the given filter and return the 'count' first
        variants. If skip the first 'skip' variants will not be regarded.

            Args:
                case_id (str): Path to a vcf file (for this adapter)
                skip (int): Skip first variants
                count (int): The number of variants to return
                filters (dict): A dictionary with filters. Currently this will
                look like: {
                    gene_list: [] (list of hgnc ids),
                    frequency: None (float),
                    cadd: None (float),
                    sv_len: None (float),
                    consequence: [] (list of consequences),
                    is_lof: None (Bool),
                    genetic_models [] (list of genetic models)
                    sv_type: List (list of sv types),
                }
            Returns:
                puzzle.constants.Results : Named tuple with variants and
                                           nr_of_variants

        """
        filters = filters or {}
        case_obj = self.case(case_id=case_id)

        limit = count + skip

        genes = set()
        if filters.get('gene_ids'):
            genes = set([gene_id.strip() for gene_id in filters['gene_ids']])

        frequency = None
        if filters.get('frequency'):
            frequency = float(filters['frequency'])

        cadd = None
        if filters.get('cadd'):
            cadd = float(filters['cadd'])

        genetic_models = None
        if filters.get('genetic_models'):
            genetic_models = set(filters['genetic_models'])

        sv_len = None
        if filters.get('sv_len'):
            sv_len = float(filters['sv_len'])

        impact_severities = None
        if filters.get('impact_severities'):
            impact_severities = set(filters['impact_severities'])

        vcf_file_path = case_obj.variant_source

        self.head = get_header(vcf_file_path)

        self.vep_header = self.head.vep_columns
        self.snpeff_header = self.head.snpeff_columns

        variants = self._get_filtered_variants(vcf_file_path, filters)

        result = []
        skip_index = 0
        for index, variant in enumerate(variants):
            index += 1
            if skip_index >= skip:
                variant_obj = self._format_variants(
                     variant=variant,
                     index=index,
                     case_obj=case_obj,
                )

                if genes and variant_obj:
                    if not set(variant_obj['gene_symbols']).intersection(genes):
                        variant_obj = None

                if impact_severities and variant_obj:
                    if not variant_obj['impact_severity'] in impact_severities:
                        variant_obj = None

                if frequency and variant_obj:
                    if variant_obj.max_freq > frequency:
                        variant_obj = None

                if cadd and variant_obj:
                    if variant_obj['cadd_score'] < cadd:
                        variant_obj = None

                if genetic_models and variant_obj:
                    models = set(variant_obj.genetic_models)
                    if not models.intersection(genetic_models):
                        variant_obj = None

                if sv_len and variant_obj:
                    if variant_obj.sv_len < sv_len:
                        variant_obj = None

                if variant_obj:
                    skip_index += 1

                    if skip_index <= limit:
                        result.append(variant_obj)
                    else:
                        break
            else:
                skip_index += 1

        return Results(result, len(result))

    def _get_filtered_variants(self, vcf_file_path, filters={}):
        """Check if variants follows the filters

            This function will try to make filters faster for the vcf adapter

            Args:
                vcf_file_path(str): Path to vcf
                filters (dict): A dictionary with filters

            Yields:
                varian_line (str): A vcf variant line
        """

        genes = set()
        consequences = set()
        sv_types = set()

        if filters.get('gene_ids'):
            genes = set([gene_id.strip() for gene_id in filters['gene_ids']])

        if filters.get('consequence'):
            consequences = set(filters['consequence'])

        if filters.get('sv_types'):
            sv_types = set(filters['sv_types'])

        logger.info("Get variants from {0}".format(vcf_file_path))

        if filters.get('range'):
            range_str = "{0}:{1}-{2}".format(
                filters['range']['chromosome'],
                filters['range']['start'],
                filters['range']['end'])

            vcf = VCF(vcf_file_path)
            handle = vcf(range_str)
        else:
            handle = VCF(vcf_file_path)

        for variant in handle:
            variant_line = str(variant)
            keep_variant = True

            if genes and keep_variant:
                keep_variant = False
                for gene in genes:
                    if "{0}".format(gene) in variant_line:
                        keep_variant = True
                        break

            if consequences and keep_variant:
                keep_variant = False
                for consequence in consequences:
                    if consequence in variant_line:
                        keep_variant = True
                        break

            if sv_types and keep_variant:
                keep_variant = False
                for sv_type in sv_types:
                    if sv_type in variant_line:
                        keep_variant = True
                        break

            if keep_variant:
                yield variant

    def _format_variants(self, variant, index, case_obj, add_all_info=False):
        """Return a Variant object

        Format variant make a variant that includes enough information for
        the variant view.
        If add_all_info then all transcripts will be parsed

        Args:
            variant (cython2.Variant): A variant object
            index (int): The index of the variant
            case_obj (puzzle.models.Case): A case object

        """
        header_line = self.head.header
        # Get the individual ids for individuals in vcf file
        vcf_individuals = set([ind_id for ind_id in self.head.individuals])

        #Create a info dict:
        info_dict = dict(variant.INFO)

        chrom = variant.CHROM
        if chrom.startswith('chr') or chrom.startswith('CHR'):
            chrom = chrom[3:]

        variant_obj = Variant(
                CHROM=chrom,
                POS=variant.POS,
                ID=variant.ID,
                REF=variant.REF,
                ALT=variant.ALT[0],
                QUAL=variant.QUAL,
                FILTER=variant.FILTER,
            )
        variant_obj._set_variant_id()

        logger.debug("Creating a variant object of variant {0}".format(
            variant_obj.variant_id))

        variant_obj.index = index
        logger.debug("Updating index to: {0}".format(
            index))

        ########### Get the coordinates for the variant ##############
        variant_obj.start = variant.start
        variant_obj.stop = variant.end

        #SV variants needs to be handeled a bit different since the can be huge
        #it would take to much power to parse all vep/snpeff entrys for these.
        if self.variant_type == 'sv':
            variant_obj.stop = int(info_dict.get('END', variant_obj.POS))
            self._add_sv_coordinates(variant_obj)
            variant_obj.sv_type = info_dict.get('SVTYPE')

            # Special for FindSV software:
            # SV specific tag for number of occurances
            occurances = info_dict.get('OCC')
            if occurances:
                logger.debug("Updating occurances to: {0}".format(
                    occurances))
                variant_obj['occurances'] = float(occurances)
                variant_obj.add_frequency('OCC', occurances)

        else:
            self._add_thousand_g(variant_obj, info_dict)
            self._add_cadd_score(variant_obj, info_dict)
            self._add_genetic_models(variant_obj, info_dict)
            self._add_transcripts(variant_obj, info_dict)
            self._add_exac(variant_obj, info_dict)
            
        self._add_hgnc_symbols(variant_obj)

        if add_all_info:
            self._add_genotype_calls(variant_obj, str(variant), case_obj)
            self._add_compounds(variant_obj, info_dict)
            self._add_gmaf(variant_obj, info_dict)
            self._add_genes(variant_obj)


        ##### Add consequences ####
        self._add_consequences(variant_obj, str(variant))
        self._add_most_severe_consequence(variant_obj)
        self._add_impact_severity(variant_obj)
        self._add_rank_score(variant_obj, info_dict)
        variant_obj.set_max_freq()
        return variant_obj




from puzzle.plugins import GeminiPlugin

from puzzle.models import DotDict

class TestFrequencies:
    
    def test_add_thousand_g(self, variant):
        adapter = GeminiPlugin()
        gemini_variant = {'aaf_1kg_all':0.01}
        adapter._add_thousand_g(variant, gemini_variant)
        
        assert variant.thousand_g == 0.01
        assert len(variant.frequencies) == 1

    def test_add_thousand_g_no_frequency(self, variant):
        adapter = GeminiPlugin()
        gemini_variant = {'aaf_1kg_all':None}
        adapter._add_thousand_g(variant, gemini_variant)
        
        assert variant.thousand_g == None
        assert len(variant.frequencies) == 0

    def test_add_exac(self, variant):
        adapter = GeminiPlugin()
        gemini_variant = {'aaf_exac_all':0.01}
        adapter._add_exac(variant, gemini_variant)
        
        assert variant.thousand_g == None
        assert len(variant.frequencies) == 1

    def test_add_exac_no_frequency(self, variant):
        adapter = GeminiPlugin()
        gemini_variant = {'aaf_exac_all':None}
        adapter._add_exac(variant, gemini_variant)
        
        assert len(variant.frequencies) == 0

    def test_add_gmaf(self, variant):
        adapter = GeminiPlugin()
        gemini_variant = {'max_aaf_all':0.01}
        adapter._add_gmaf(variant, gemini_variant)
        
        assert len(variant.frequencies) == 0
        assert variant.max_freq == 0.01

    def test_add_gmaf_no_frequency(self, variant):
        adapter = GeminiPlugin()
        gemini_variant = {'max_aaf_all':None}
        adapter._add_gmaf(variant, gemini_variant)
        
        assert len(variant.frequencies) == 0
        assert variant.max_freq == None

class TestGenotypes:
    
    def test_add_genotypes(self, variant):
        adapter = GeminiPlugin()
        gemini_variant = {
            'gts': ['G/A', 'G/A', './.', './.', './.'],
            'gt_depths': [17, 19, -1, -1, -1],
            'gt_ref_depths': [10, 13, -1, -1, -1],
            'gt_alt_depths': [7, 6, -1, -1, -1],
            'gt_quals': [99, 99, -1, -1, -1],
            
        }
        ind = DotDict()
        ind.ind_index = 0
        ind.ind_id = '1'
        ind.phenotype = 2
        ind_objs = [ind]

        adapter._add_genotypes(variant, gemini_variant, 'dummy', ind_objs)
        
        genotype = variant.individuals[0]

        assert genotype.sample_id == ind.ind_id
        assert genotype.sample_id == ind.ind_id
        assert genotype.genotype == 'G/A'
        assert genotype.case_id == 'dummy'
        assert genotype.phenotype == ind.phenotype
        assert genotype.ref_depth == 10
        assert genotype.alt_depth == 7
        assert genotype.depth == 17
        assert genotype.genotype_quality == 99

class TestTranscripts:
    
    def test_add_transcripts(self, gemini_case_obj, variant):
        adapter = GeminiPlugin()
        adapter.add_case(gemini_case_obj)
        adapter.db = gemini_case_obj.variant_source
    
        gemini_variant = {'variant_id': 1}
        adapter._add_transcripts(variant, gemini_variant)
        
        transcripts = variant.transcripts
        
        assert len(transcripts) == 2

        first_transcript = transcripts[0]

        assert first_transcript.transcript_id == 'ENST00000370383'
        assert first_transcript.hgnc_symbol == 'EPHX4'
        assert first_transcript.consequence == 'missense_variant'
        assert first_transcript.ensembl_id == None
        assert first_transcript.biotype == 'protein_coding'
        assert first_transcript.sift == 'deleterious'
        assert first_transcript.polyphen == 'probably_damaging'
    


import os

import logging
import click

from . import (base, root)

from puzzle.plugins import SqlStore

logger = logging.getLogger(__name__)


@base.command()
@root
@click.pass_context
def cases(ctx, root):
    """
    Show all cases in the database.

    If no database was found run puzzle init first.
    """
    root = root or ctx.obj.get('root') or os.path.expanduser("~/.puzzle")

    if os.path.isfile(root):
        logger.error("'root' can't be a file")
        ctx.abort()

    logger.info("Root directory is: {}".format(root))

    db_path = os.path.join(root, 'puzzle_db.sqlite3')
    logger.info("db path is: {}".format(db_path))

    if not os.path.exists(db_path):
        logger.warn("database not initialized, run 'puzzle init'")
        ctx.abort()

    store = SqlStore(db_path)
    for case in store.cases():
        click.echo(case)



from puzzle.utils import (get_gene_info, get_gene_symbols)


class GeneExtras(object):
    """Methods for genes"""
    
    def _add_hgnc_symbols(self, variant_obj):
        """Add hgnc symbols to the variant
        
            If there are transcripts use the symbols found here,
            otherwise use phizz to get the gene ids.
        """
        hgnc_symbols = set()
        if variant_obj.transcripts:
            for transcript in variant_obj.transcripts:
                if transcript.hgnc_symbol:
                    hgnc_symbols.add(transcript.hgnc_symbol)
        else:
            chrom = variant_obj.CHROM
            start = variant_obj.start
            stop = variant_obj.stop
        
            hgnc_symbols = get_gene_symbols(chrom, start, stop)
        #Make unique ids
        variant_obj.gene_symbols = list(hgnc_symbols)
        
    def _add_genes(self, variant_obj):
        """Add the Gene objects for a variant"""
        genes = []
        ensembl_ids = []
        hgnc_symbols = []
        
        if variant_obj.transcripts:
            for transcript in variant_obj.transcripts:
                if transcript.ensembl_id:
                    ensembl_ids.append(transcript.ensembl_id)
                if transcript.hgnc_symbol:
                    hgnc_symbols.append(transcript.hgnc_symbol)

        else:
            hgnc_symbols = variant_obj.gene_symbols
        
        genes = get_gene_info(
                        ensembl_ids=ensembl_ids, 
                        hgnc_symbols=hgnc_symbols
                        )
            
        for gene in genes:
            variant_obj.add_gene(gene)
        


##TODO mock a gemini db
import pytest
from puzzle.plugins import GeminiPlugin

from sqlite3 import DatabaseError

def test_setup_no_db():
    """Test to initialize a gemini database without any file"""

    adapter = GeminiPlugin()
    assert adapter.db == None
    
    assert adapter.variant_type == 'snv'
    
    assert adapter.filters.can_filter_gene == True
    assert adapter.filters.can_filter_inheritance == False

def test_setup_with_variant_type(gemini_path):
    adapter = GeminiPlugin('sv')
    assert adapter.variant_type == 'sv'

def test_check_gemini_db(gemini_path, vcf_file):
    adapter = GeminiPlugin()
    adapter.db = gemini_path
    assert adapter.test_gemini_db()
    
    adapter.db = vcf_file
    with pytest.raises(DatabaseError):
        adapter.test_gemini_db()



# -*- coding: utf-8 -*-
"""
test_puzzle
----------------------------------

Tests for `puzzle` module.
"""
import pytest
import puzzle


class TestPuzzle(object):
    @classmethod
    def set_up(self):
        pass

    def test_something(self):
        pass

    @classmethod
    def tear_down(self):
        pass



# -*- coding: utf-8 -*-
import os

from flask import (abort, Blueprint, current_app as app, render_template,
                   redirect, request, url_for, make_response,
                   send_from_directory, flash)
from werkzeug import secure_filename

from puzzle.models.sql import Case

BP_NAME = __name__.split('.')[-2]
blueprint = Blueprint(BP_NAME, __name__, template_folder='templates',
                      static_folder='static',
                      static_url_path="/{}/static".format(BP_NAME))


@blueprint.route('/')
def index():
    """Show the landing page."""
    gene_lists = app.db.gene_lists() if app.config['STORE_ENABLED'] else []
    queries = app.db.gemini_queries() if app.config['STORE_ENABLED'] else []

    case_groups = {}
    for case in app.db.cases():
        key = (case.variant_source, case.variant_type, case.variant_mode)
        if key not in case_groups:
            case_groups[key] = []
        case_groups[key].append(case)

    return render_template('index.html', case_groups=case_groups,
                           gene_lists=gene_lists, queries=queries)


@blueprint.route('/cases/<case_id>')
def case(case_id):
    """Show the overview for a case."""
    case_obj = app.db.case(case_id)
    return render_template('case.html', case=case_obj, case_id=case_id)


@blueprint.route('/phenotypes', methods=['POST'])
def phenotypes():
    """Add phenotype(s) to the case model."""
    ind_id = request.form['ind_id']
    phenotype_id = request.form['phenotype_id']

    if not phenotype_id:
        return abort(500, 'no phenotype_id submitted')

    ind_obj = app.db.individual(ind_id)
    try:
        added_terms = app.db.add_phenotype(ind_obj, phenotype_id)
        if added_terms is None:
            flash("Term with id {} was not found".format(phenotype_id),
                  'danger')
        elif added_terms == []:
            flash("Term with id {} was already added".format(phenotype_id),
                  'warning')
    except RuntimeError as error:
        return abort(500, error.message)

    return redirect(request.referrer)


@blueprint.route('/phenotypes/<phenotype_id>/delete', methods=['POST'])
def delete_phenotype(phenotype_id):
    """Delete phenotype from an individual."""
    ind_id = request.form['ind_id']
    ind_obj = app.db.individual(ind_id)
    try:
        app.db.remove_phenotype(ind_obj, phenotype_id)
    except RuntimeError as error:
        return abort(500, error.message)
    return redirect(request.referrer)


@blueprint.route('/genelists', methods=['POST'])
@blueprint.route('/genelists/<list_id>', methods=['GET', 'POST'])
def gene_list(list_id=None):
    """Display or add a gene list."""
    all_case_ids = [case.case_id for case in app.db.cases()]
    if list_id:
        genelist_obj = app.db.gene_list(list_id)
        case_ids = [case.case_id for case in app.db.cases()
                    if case not in genelist_obj.cases]
        if genelist_obj is None:
            return abort(404, "gene list not found: {}".format(list_id))

    if 'download' in request.args:
        response = make_response('\n'.join(genelist_obj.gene_ids))
        filename = secure_filename("{}.txt".format(genelist_obj.list_id))
        header = "attachment; filename={}".format(filename)
        response.headers['Content-Disposition'] = header
        return response

    if request.method == 'POST':
        if list_id:
            # link a case to the gene list
            case_ids = request.form.getlist('case_id')
            for case_id in case_ids:
                case_obj = app.db.case(case_id)
                if case_obj not in genelist_obj.cases:
                    genelist_obj.cases.append(case_obj)
                    app.db.save()
        else:
            # upload a new gene list
            req_file = request.files['file']
            new_listid = (request.form['list_id'] or
                          secure_filename(req_file.filename))

            if app.db.gene_list(new_listid):
                return abort(500, 'Please provide a unique list name')

            if not req_file:
                return abort(500, 'Please provide a file for upload')

            gene_ids = [line for line in req_file.stream
                        if not line.startswith('#')]
            genelist_obj = app.db.add_genelist(new_listid, gene_ids)
            case_ids = all_case_ids

    return render_template('gene_list.html', gene_list=genelist_obj,
                           case_ids=case_ids)


@blueprint.route('/genelists/delete/<list_id>', methods=['POST'])
@blueprint.route('/genelists/delete/<list_id>/<case_id>', methods=['POST'])
def delete_genelist(list_id, case_id=None):
    """Delete a whole gene list with links to cases or a link."""
    if case_id:
        # unlink a case from a gene list
        case_obj = app.db.case(case_id)
        app.db.remove_genelist(list_id, case_obj=case_obj)
        return redirect(request.referrer)
    else:
        # remove the whole gene list
        app.db.remove_genelist(list_id)
        return redirect(url_for('.index'))


@blueprint.route('/resources', methods=['POST'])
def resources():
    """Upload a new resource for an individual."""
    ind_id = request.form['ind_id']

    upload_dir = os.path.abspath(app.config['UPLOAD_DIR'])
    req_file = request.files['file']
    filename = secure_filename(req_file.filename)
    file_path = os.path.join(upload_dir, filename)
    name = request.form['name'] or filename
    req_file.save(file_path)

    ind_obj = app.db.individual(ind_id)
    app.db.add_resource(name, file_path, ind_obj)
    return redirect(request.referrer)


@blueprint.route('/resources/<resource_id>')
def resource(resource_id):
    """Show a resource."""
    resource_obj = app.db.resource(resource_id)

    if 'raw' in request.args:
        return send_from_directory(os.path.dirname(resource_obj.path),
                                   os.path.basename(resource_obj.path))

    return render_template('resource.html', resource=resource_obj)


@blueprint.route('/resource/delete/<resource_id>', methods=['POST'])
def delete_resource(resource_id):
    """Delete a resource."""
    resource_obj = app.db.resource(resource_id)
    try:
        os.remove(resource_obj.path)
    except OSError as err:
        app.logger.debug(err.message)

    app.db.delete_resource(resource_id)
    return redirect(request.referrer)


@blueprint.route('/<case_id>/comments', methods=['POST'])
def comments(case_id):
    """Upload a new comment."""
    text = request.form['text']
    variant_id = request.form.get('variant_id')
    username = request.form.get('username')
    case_obj = app.db.case(case_id)
    app.db.add_comment(case_obj, text, variant_id=variant_id, username=username)
    return redirect(request.referrer)


@blueprint.route('/comments/<comment_id>', methods=['POST'])
def delete_comment(comment_id):
    """Delete a comment."""
    app.db.delete_comment(comment_id)
    return redirect(request.referrer)


@blueprint.route('/individuals')
def individuals():
    """Show an overview of all individuals."""
    individual_objs = app.db.individuals()
    return render_template('individuals.html', individuals=individual_objs)


@blueprint.route('/individuals/<ind_id>')
def individual(ind_id):
    """Show details for a specific individual."""
    individual_obj = app.db.individual(ind_id)
    return render_template('individual.html', individual=individual_obj)


@blueprint.route('/<case_id>/synopsis', methods=['POST'])
def synopsis(case_id):
    """Update the case synopsis."""
    text = request.form['text']
    case_obj = app.db.case(case_id)
    app.db.update_synopsis(case_obj, text)
    return redirect(request.referrer)


@blueprint.route('/cases', methods=['POST'])
def add_case():
    """Make a new case out of a list of individuals."""
    ind_ids = request.form.getlist('ind_id')
    case_id = request.form['case_id']
    source = request.form['source']
    variant_type = request.form['type']

    if len(ind_ids) == 0:
        return abort(400, "must add at least one member of case")

    # only GEMINI supported
    new_case = Case(case_id=case_id, name=case_id, variant_source=source,
                    variant_type=variant_type, variant_mode='gemini')

    # Add individuals to the correct case
    for ind_id in ind_ids:
        ind_obj = app.db.individual(ind_id)
        new_case.individuals.append(ind_obj)

    app.db.session.add(new_case)
    app.db.save()

    return redirect(url_for('.case', case_id=new_case.name))



# -*- coding: utf-8 -*-
import logging

from .dotdict import DotDict

logger = logging.getLogger(__name__)


class Transcript(DotDict):
    """Class that holds information about a transcript"""
    def __init__(self, hgnc_symbol, transcript_id, consequence, ensembl_id=None, biotype=None, 
                strand=None, sift=None, polyphen=None, exon=None, HGVSc=None, 
                HGVSp=None, GMAF=None, ExAC_MAF=None):
        super(Transcript, self).__init__(
            hgnc_symbol=hgnc_symbol, transcript_id=transcript_id, biotype=biotype,
            consequence=consequence, ensembl_id=ensembl_id, strand=strand, sift=sift,
            polyphen=polyphen, exon=exon, HGVSc=HGVSc, HGVSp=HGVSp, GMAF=GMAF, 
            ExAC_MAF=ExAC_MAF)



import logging

logger = logging.getLogger(__name__)

class FrequenciesExtras(object):
    """Methods for adding frequencies"""
    
    def _add_thousand_g(self, variant_obj, info_dict):
        """Add the thousand genomes frequency
        
        Args:
            variant_obj (puzzle.models.Variant)
            info_dict (dict): A info dictionary
        
        """
        thousand_g = info_dict.get('1000GAF')
        if thousand_g:
            logger.debug("Updating thousand_g to: {0}".format(
                thousand_g))
            variant_obj.thousand_g = float(thousand_g)
            variant_obj.add_frequency('1000GAF', variant_obj.get('thousand_g'))
    
    def _add_gmaf(self, variant_obj, info_dict):
        """Add the gmaf frequency
        
        Args:
            variant_obj (puzzle.models.Variant)
            info_dict (dict): A info dictionary
        
        """
        ##TODO search for max freq in info dict
        for transcript in variant_obj.transcripts:
            gmaf_raw = transcript.GMAF
            if gmaf_raw:
                gmaf = float(gmaf_raw.split(':')[-1])
                variant_obj.add_frequency('GMAF', gmaf)
                
                if not variant_obj.thousand_g:
                    variant_obj.thousand_g = gmaf
    
    def _add_exac(self, variant_obj, info_dict):
        """Add the gmaf frequency
        
        Args:
            variant_obj (puzzle.models.Variant)
            info_dict (dict): A info dictionary
        
        """
        exac = None
        exac_keys = ['ExAC', 'EXAC', 'ExACAF', 'EXACAF']
        for key in exac_keys:
            if info_dict.get(key):
                exac = float(info_dict[key])
        #If not found in vcf search transcripts
        if not exac:
            for transcript in variant_obj.transcripts:
                exac_raw = transcript.ExAC_MAF
                if exac_raw:
                    exac = float(exac_raw.split(':')[-1])
        
        if exac:
            variant_obj.add_frequency('ExAC', exac)
    


# -*- coding: utf-8 -*-
import logging

from . import DotDict

logger = logging.getLogger(__name__)


class Compound(DotDict):
    """Class that holds information about a compound variant"""
    def __init__(self, variant_id, combined_score=None):
        super(Compound, self).__init__(variant_id=variant_id,
                                       combined_score=combined_score)



import logging

from cyvcf2 import VCF
try:
    from gemini import GeminiQuery
except ImportError:
    pass

logger = logging.getLogger(__name__)

def get_file_type(variant_source):
    """Check what kind of file variant source is
    
        Args:
            variant_source (str): Path to variant source
        
        Returns:
            file_type (str): 'vcf', 'gemini' or 'unknown'
    """
    file_type = 'unknown'
    valid_vcf_suffixes = ('.vcf', '.vcf.gz')
    if variant_source:
        logger.debug("Check file type with file: {0}".format(variant_source))
        if variant_source.endswith('.db'):
            file_type = 'gemini'
            logger.debug("File {0} is a gemini database".format(variant_source))
        elif variant_source.endswith(valid_vcf_suffixes):
            file_type = 'vcf'
            logger.debug("File {0} is a vcf".format(variant_source))
        else:
            logger.debug("File is in a unknown format")
    
    return file_type

def get_variant_type(variant_source):
    """Try to find out what type of variants that exists in a variant source
    
        Args:
            variant_source (str): Path to variant source
            source_mode (str): 'vcf' or 'gemini'
        
        Returns:
            variant_type (str): 'sv' or 'snv'
    """
    file_type = get_file_type(variant_source)
    variant_type = 'sv'
    if file_type == 'vcf':
        variants = VCF(variant_source)
    elif file_type == 'gemini':
        variants = GeminiQuery(variant_source)
        gemini_query = "SELECT * from variants"
        variants.run(gemini_query)
    # Check 1000 first variants, if anyone is a snv we set the variant_type
    # to 'snv'
    for i,variant in enumerate(variants):
        if file_type == 'vcf':
            if variant.is_snp:
                variant_type = 'snv'
        elif file_type == 'gemini':
            if variant['type'] == 'snp':
                variant_type = 'snv'
            
        if i > 1000:
            break
    
    return variant_type
                    
            
    



from .utils import (version, family_file, family_type, verbose, root, 
 mode, variant_type, phenomizer)

from .base import base
from .init import init
from .delete import delete
from .load import load
from .individuals import individuals
from .cases import cases
from .view import view


# -*- coding: utf-8 -*-


class PedigreeHumanMixin(object):
    @property
    def sex_human(self):
        """Return a human readable string for the sex."""
        sex_string = self.sex
        if sex_string == '1':
            return 'male'
        elif sex_string == '2':
            return 'female'
        else:
            return 'unknown'

    @property
    def is_affected(self):
        """Boolean for telling if the sample is affected."""
        phenotype = self.phenotype
        if phenotype == '1':
            return False
        elif phenotype == '2':
            return True
        else:
            return False



import logging

from gemini import GeminiQuery

from puzzle.plugins import BaseVariantMixin
from puzzle.plugins.constants import Results

from puzzle.models import (Compound, Variant, Gene, Genotype, Transcript,)
from puzzle.utils import (get_most_severe_consequence, get_omim_number,
                          get_cytoband_coord)

from . import VariantExtras

logger = logging.getLogger(__name__)

class VariantMixin(BaseVariantMixin, VariantExtras):
    """Class to store variant specific functions for gemini plugin"""


    def build_gemini_query(self, query, extra_info):
        """Append sql to a gemini query

        Args:
            query(str): The gemini query
            extra_info(str): The text that should be added

        Return:
            extended_query(str)
        """
        if 'WHERE' in query:
            return "{0} AND {1}".format(query, extra_info)
        else:
            return "{0} WHERE {1}".format(query, extra_info)

    def variants(self, case_id, skip=0, count=1000, filters=None):
        """Return count variants for a case.

        This function needs to have different behaviours based on what is asked
        for. It should allways try to give minimal information back to improve
        on speed. For example, if consequences are not asked for we will not
        build all transcripts. If not sv variants we will not build sv
        coordinates.
        So the minimal case is to just show what is asked for in the variants
        interface.

            Args:
                case_id (str): A gemini db
                skip (int): Skip first variants
                count (int): The number of variants to return
                filters (dict): A dictionary with filters. Currently this will
                look like: {
                    gene_list: [] (list of hgnc ids),
                    frequency: None (float),
                    cadd: None (float),
                    consequence: [] (list of consequences),
                    impact_severities: [] (list of consequences),
                    genetic_models [] (list of genetic models)
                }
            Returns:
                puzzle.constants.Results : Named tuple with variants and
                                           nr_of_variants

        """
        filters = filters or {}
        logger.debug("Looking for variants in {0}".format(case_id))

        limit = count + skip

        gemini_query = filters.get('gemini_query') or "SELECT * from variants v"

        any_filter = False

        if filters.get('frequency'):
            frequency = filters['frequency']

            extra_info = "(v.max_aaf_all < {0} or v.max_aaf_all is"\
                         " Null)".format(frequency)
            gemini_query = self.build_gemini_query(gemini_query, extra_info)

        if filters.get('cadd'):
            cadd_score = filters['cadd']

            extra_info = "(v.cadd_scaled > {0})".format(cadd_score)
            gemini_query = self.build_gemini_query(gemini_query, extra_info)

        if filters.get('gene_ids'):
            gene_list = [gene_id.strip() for gene_id in filters['gene_ids']]

            gene_string = "v.gene in ("
            for index, gene_id in enumerate(gene_list):
                if index == 0:
                    gene_string += "'{0}'".format(gene_id)
                else:
                    gene_string += ", '{0}'".format(gene_id)
            gene_string += ")"

            gemini_query = self.build_gemini_query(gemini_query, gene_string)

        if filters.get('range'):
            chrom = filters['range']['chromosome']
            if not chrom.startswith('chr'):
                chrom = "chr{0}".format(chrom)

            range_string = "v.chrom = '{0}' AND "\
                           "((v.start BETWEEN {1} AND {2}) OR "\
                           "(v.end BETWEEN {1} AND {2}))".format(
                               chrom,
                               filters['range']['start'],
                               filters['range']['end']
                           )
            gemini_query = self.build_gemini_query(gemini_query, range_string)

        filtered_variants = self._variants(
            case_id=case_id,
            gemini_query=gemini_query,
        )

        if filters.get('consequence'):
            consequences = set(filters['consequence'])

            filtered_variants = (variant for variant in filtered_variants if
                set(variant.consequences).intersection(consequences))

        if filters.get('impact_severities'):
            severities = set([severity.strip()
                    for severity in filters['impact_severities']])
            new_filtered_variants = []
            filtered_variants = (variant for variant in filtered_variants if
                set([variant.impact_severity]).intersection(severities))

        if filters.get('sv_len'):
            sv_len = int(filters['sv_len'])
            filtered_variants = (variant for variant in filtered_variants if
                variant.sv_len >= sv_len)

        variants = []
        for index, variant_obj in enumerate(filtered_variants):
            if index >= skip:
                if index < limit:
                    variants.append(variant_obj)
                else:
                    break

        return Results(variants, len(variants))

    def variant(self, case_id, variant_id):
        """Return a specific variant.

            We solve this by building a gemini query and send it to _variants

            Args:
                case_id (str): Path to a gemini database
                variant_id (int): A gemini variant id

            Returns:
                variant_obj (dict): A puzzle variant

        """
        #Use the gemini id for fast lookup
        variant_id = int(variant_id)
        gemini_query = "SELECT * from variants WHERE variant_id = {0}".format(
            variant_id
        )

        individuals = []
        # Get the individuals for the case
        case_obj = self.case(case_id)
        for individual in case_obj.individuals:
            individuals.append(individual)

        self.db = case_obj.variant_source
        self.variant_type = case_obj.variant_type

        gq = GeminiQuery(self.db)
        gq.run(gemini_query)

        for gemini_variant in gq:
            variant = self._format_variant(
                case_id=case_id,
                gemini_variant=gemini_variant,
                individual_objs=individuals,
                index=gemini_variant['variant_id'],
                add_all_info = True
            )
            return variant

        return None

    def _variants(self, case_id, gemini_query):
        """Return variants found in the gemini database

            Args:
                case_id (str): The case for which we want to see information
                gemini_query (str): What variants should be chosen
                filters (dict): A dictionary with filters

            Yields:
                variant_obj (dict): A Variant formatted dictionary
        """
        individuals = []
        # Get the individuals for the case
        case_obj = self.case(case_id)
        for individual in case_obj.individuals:
            individuals.append(individual)

        self.db = case_obj.variant_source
        self.variant_type = case_obj.variant_type

        gq = GeminiQuery(self.db)

        gq.run(gemini_query)

        index = 0
        for gemini_variant in gq:
            variant = None

            # Check if variant is non ref in the individuals
            is_variant = self._is_variant(gemini_variant, individuals)

            if self.variant_type == 'snv' and not is_variant:
                variant = None

            else:
                index += 1
                logger.debug("Updating index to: {0}".format(index))
                variant = self._format_variant(
                        case_id=case_id,
                        gemini_variant=gemini_variant,
                        individual_objs=individuals,
                        index=index
                        )

            if variant:

                yield variant

    def _format_variant(self, case_id, gemini_variant, individual_objs,
                        index=0, add_all_info=False):
        """Make a puzzle variant from a gemini variant

            Args:
                case_id (str): related case id
                gemini_variant (GeminiQueryRow): The gemini variant
                individual_objs (list(dict)): A list of Individuals
                index(int): The index of the variant

            Returns:
                variant (dict): A Variant object
        """
        chrom = gemini_variant['chrom']
        if chrom.startswith('chr') or chrom.startswith('CHR'):
            chrom = chrom[3:]

        variant_dict = {
            'CHROM':chrom,
            'POS':str(gemini_variant['start']),
            'ID':gemini_variant['rs_ids'],
            'REF':gemini_variant['ref'],
            'ALT':gemini_variant['alt'],
            'QUAL':gemini_variant['qual'],
            'FILTER':gemini_variant['filter']
        }

        variant = Variant(**variant_dict)

        # Use the gemini id for fast search
        variant.update_variant_id(gemini_variant['variant_id'])
        logger.debug("Creating a variant object of variant {0}".format(
            variant.variant_id))

        variant['index'] = index

        # Add the most severe consequence
        self._add_most_severe_consequence(variant, gemini_variant)

        #Add the impact severity
        self._add_impact_severity(variant, gemini_variant)
        ### POSITON ANNOATTIONS ###
        variant.start = int(gemini_variant['start'])
        variant.stop = int(gemini_variant['end'])

        #Add the sv specific coordinates
        if self.variant_type == 'sv':
            variant.sv_type = gemini_variant['sub_type']
            variant.stop = int(gemini_variant['end'])
            self._add_sv_coordinates(variant)

        else:
            ### Consequence and region annotations
            #Add the transcript information
            self._add_transcripts(variant, gemini_variant)
            self._add_thousand_g(variant, gemini_variant)
            self._add_exac(variant, gemini_variant)
            self._add_gmaf(variant, gemini_variant)
            #### Check the impact annotations ####
            if gemini_variant['cadd_scaled']:
                variant.cadd_score = gemini_variant['cadd_scaled']

            # We use the prediction in text
            polyphen = gemini_variant['polyphen_pred']
            if polyphen:
                variant.add_severity('Polyphen', polyphen)

            # We use the prediction in text
            sift = gemini_variant['sift_pred']
            if sift:
                variant.add_severity('SIFT', sift)

        #Add the genes based on the hgnc symbols
        self._add_hgnc_symbols(variant)
        if self.variant_type == 'snv':
            self._add_genes(variant)

        self._add_consequences(variant)

        ### GENOTYPE ANNOATTIONS ###
        #Get the genotype info
        if add_all_info:
            self._add_genotypes(variant, gemini_variant, case_id, individual_objs)
            if self.variant_type == 'sv':
                self._add_genes(variant)

        return variant

    def _is_variant(self, gemini_variant, ind_objs):
        """Check if the variant is a variation in any of the individuals

        Args:
            gemini_variant (GeminiQueryRow): The gemini variant
            ind_objs (list(puzzle.models.individual)): A list of individuals to check

        Returns:
            bool : If any of the individuals has the variant
        """

        indexes = (ind.ind_index for ind in ind_objs)
        #Check if any individual have a heterozygous or homozygous variant call
        for index in indexes:
            gt_call = gemini_variant['gt_types'][index]
            if (gt_call == 1 or gt_call == 3):
                return True

        return False



# -*- coding: utf-8 -*-


def test_gemini_query(test_db):
    only_query = test_db.gemini_queries()[0]
    query_obj = test_db.gemini_query(only_query.id)
    assert query_obj.name == only_query.name


def test_gemini_queries(test_db):
    query_objs = test_db.gemini_queries()
    assert query_objs.count() == 1


def test_add_gemini_query(test_db):
    name = 'Bruce Lee Query'
    query = "SELECT * FROM variants WHERE 'lee' > 'chuck'"
    new_query = test_db.add_gemini_query(name, query)
    assert isinstance(new_query.id, int)


def test_delete_query(test_db):
    only_query = test_db.gemini_queries()[0]
    assert test_db.gemini_query(only_query.id)
    test_db.delete_gemini_query(only_query.id)
    assert test_db.gemini_query(only_query.id) is None



# -*- coding: utf-8 -*-
from sqlalchemy import (Column, ForeignKey, Integer, String, UniqueConstraint,
                        Text)
from sqlalchemy.orm import relationship

from .models import BASE


class Suspect(BASE):

    """Represent a list of suspect variants."""

    __tablename__ = "suspect"
    __table_args__ = (UniqueConstraint('case_id', 'variant_id',
                                       name='_case_variant_uc'),)

    id = Column(Integer, primary_key=True)
    variant_id = Column(Text, nullable=False)
    name = Column(String(128))

    case_id = Column(Integer, ForeignKey('case.id'), nullable=False)
    case = relationship('Case', backref=('suspects'))

    def __repr__(self):
        return ("Suspect(case_id={this.case_id} variant_id={this.variant_id})"
                .format(this=self))



# -*- coding: utf-8 -*-
import logging

from puzzle.plugins import BaseCaseMixin
from puzzle.models import Case as BaseCase, Individual as BaseIndividual
from puzzle.models.sql import Case, Individual

logger = logging.getLogger(__name__)


class CaseMixin(BaseCaseMixin):

    """Handle case related methods."""

    def add_case(self, case_obj, vtype='snv', mode='vcf', ped_svg=None):
        """Load a case with individuals.

        Args:
            case_obj (puzzle.models.Case): initialized case model
        """
        new_case = Case(case_id=case_obj.case_id,
                        name=case_obj.name,
                        variant_source=case_obj.variant_source,
                        variant_type=vtype,
                        variant_mode=mode,
                        pedigree=ped_svg,
                        compressed=case_obj.compressed,
                        tabix_index=case_obj.tabix_index)

        # build individuals
        inds = [Individual(
            ind_id=ind.ind_id,
            name=ind.name,
            mother=ind.mother,
            father=ind.father,
            sex=ind.sex,
            phenotype=ind.phenotype,
            ind_index=ind.ind_index,
            variant_source=ind.variant_source,
            bam_path=ind.bam_path,
        ) for ind in case_obj.individuals]
        new_case.individuals = inds
        
        if self.case(new_case.case_id):
            logger.warning("Case already exists in database!")
        else:
            self.session.add(new_case)
            self.save()
        return new_case

    def delete_case(self, case_obj):
        """Delete a case from the database

        Args:
            case_obj (puzzle.models.Case): initialized case model
        """
        for ind_obj in case_obj.individuals:
            self.delete_individual(ind_obj)
        logger.info("Deleting case {0} from database".format(case_obj.case_id))
        self.session.delete(case_obj)
        self.save()
        return case_obj

    def delete_individual(self, ind_obj):
        """Delete a case from the database

        Args:
            ind_obj (puzzle.models.Individual): initialized individual model
        """
        logger.info("Deleting individual {0} from database"
                    .format(ind_obj.ind_id))
        self.session.delete(ind_obj)
        self.save()
        return ind_obj

    def case(self, case_id):
        """Fetch a case from the database."""
        case_obj = self.query(Case).filter_by(case_id=case_id).first()
        return case_obj

    def individual(self, ind_id):
        """Fetch a case from the database."""
        ind_obj = self.query(Individual).filter_by(ind_id=ind_id).first()
        if ind_obj is None:
            ind_obj = BaseIndividual(ind_id='unknown')
        return ind_obj

    def cases(self):
        """Fetch all cases from the database."""
        return self.query(Case)

    def individuals(self, ind_ids=None):
        """Fetch all individuals from the database."""
        query = self.query(Individual)
        if ind_ids:
            query = query.filter(Individual.ind_id.in_(ind_ids))
        return query



# -*- coding: utf-8 -*-
import os
import logging

from path import path

from puzzle.plugins import Plugin
from puzzle.models import DotDict
from puzzle.utils import get_cases
from .mixins import VariantMixin, CaseMixin

logger = logging.getLogger(__name__)


class VcfPlugin(VariantMixin, CaseMixin, Plugin):
    """docstring for Plugin"""

    def __init__(self, variant_type='snv'):
        """Initialize a vcf adapter.

            When instansiating all cases are found.

            Args:
                variant_type(str) : 'snv' or 'sv'
        """
        super(VcfPlugin, self).__init__()

        self.individual_objs = []
        self.case_objs = []

        self.variant_type = variant_type
        logger.info("Setting variant type to {0}".format(variant_type))

        self.variant_columns = ['CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER']
        
        self.head = None
        self.vep_header = None
        self.snpeff_header = None
        
        
        self.filters.can_filter_gene = True
        self.filters.can_filter_frequency = True
        self.filters.can_filter_cadd = True
        self.filters.can_filter_consequence = True
        self.filters.can_filter_impact_severity = True
        self.filters.can_filter_sv = True
        self.filters.can_filter_sv_len = True
        self.filters.can_filter_inheritance = True

    def init_app(self, app):
        """Initialize plugin via Flask."""
        pass




# -*- coding: utf-8 -*-
import logging

from puzzle.models.sql import GeminiQuery

logger = logging.getLogger(__name__)


class GeminiActions(object):
    def gemini_query(self, query_id):
        """Return a gemini query

        Args:
            name (str)
        """
        logger.debug("Looking for query with id {0}".format(query_id))
        return self.query(GeminiQuery).filter_by(id=query_id).first()

    def gemini_queries(self):
        """Return all gemini queries"""
        return self.query(GeminiQuery)

    def add_gemini_query(self, name, query):
        """Add a user defined gemini query

        Args:
            name (str)
            query (str)
        """
        logger.info("Adding query {0} with text {1}".format(name, query))
        new_query = GeminiQuery(name=name, query=query)
        self.session.add(new_query)
        self.save()
        return new_query

    def delete_gemini_query(self, query_id):
        """Delete a gemini query

        Args:
            name (str)
        """
        query_obj = self.gemini_query(query_id)
        logger.debug("Delete query: {0}".format(query_obj.name_query))
        self.session.delete(query_obj)
        self.save()



import codecs
import os

import logging
import click
import yaml

from . import (base, root, phenomizer)

from puzzle.plugins import SqlStore
from puzzle.constants import PUZZLE_CONFIG_PATH

logger = logging.getLogger(__name__)

@base.command()
@click.option('--reset', 
    is_flag=True,
    help="Wipe the database and initialize a new one"
)
@root
@phenomizer
@click.pass_context
def init(ctx, reset, root, phenomizer):
    """Initialize a database that store metadata

        Check if "root" dir exists, otherwise create the directory and
        build the database. If a database already exists, do nothing.

    """
    configs = {}
    if root is None:
        root = ctx.obj.get('root') or os.path.expanduser("~/.puzzle")
    configs['root'] = root
    
    if os.path.isfile(root):
        logger.error("'root' can't be a file")
        ctx.abort()

    logger.info("Root directory is: {}".format(root))

    db_path = os.path.join(root, 'puzzle_db.sqlite3')
    logger.info("db path is: {}".format(db_path))
    
    resource_dir = os.path.join(root, 'resources')
    logger.info("resource dir is: {}".format(resource_dir))

    if os.path.exists(resource_dir):
        logger.debug("Found puzzle directory: {0}".format(root))
        if os.path.exists(resource_dir) and not reset:
            logger.warning("Puzzle db already in place")
            ctx.abort()
    else:
        logger.info("Create directory: {0}".format(resource_dir))
        os.makedirs(resource_dir)
        logger.debug('Directory created')

    logger.debug('Connect to database and create tables')
    store = SqlStore(db_path)
    store.set_up(reset=reset)
    
    if phenomizer:
        phenomizer = [str(term) for term in phenomizer]
        configs['phenomizer_auth'] = phenomizer
    
    if not ctx.obj.get('config_path'):
        logger.info("Creating puzzle config file in {0}".format(PUZZLE_CONFIG_PATH))
        with codecs.open(PUZZLE_CONFIG_PATH, 'w', encoding='utf-8') as f:
            f.write(yaml.dump(configs))
        logger.debug("Config created")
            




# -*- coding: utf-8 -*-
from .mixins import PedigreeHumanMixin
from .dotdict import DotDict


class Individual(DotDict, PedigreeHumanMixin):
    """Individual representation."""
    def __init__(self, ind_id, case_id=None, mother=None,
                 father=None, sex=None, phenotype=None, ind_index=None,
                 variant_source=None, bam_path=None):
        """Construct a individual object

            Args:
                ind_id (str): The individual id
                case_id (str): What case it belongs to
                mother (str): The mother id
                father (str): The father id
                sex (str): Sex in ped format
                phenotype (str): Phenotype in ped format
                ind_index (int): Either the column in the vcf or what position in
                             the gemini database the individual represents.
                variant_source (str): Path to source (vcf file)
                bam_path (str): Path to bamfiles (vcf file)
        """
        super(Individual, self).__init__(
            ind_id=ind_id,
            name=ind_id,
            case_id=case_id,
            mother=mother,
            father=father,
            sex=sex,
            phenotype=phenotype,
            ind_index=ind_index,
            variant_source=variant_source,
            bam_path=bam_path
        )
